from typing import Callable, Union

# Representing subject=None with Optional[str] doesn't work.
ManFn = Union[Callable[[str], str], Callable[[], str]]
# from typing import runtime_checkable


# @runtime_checkable
# class ManFn(Callable[[str], str]):
# class ManFn(Callable,FunctionType):
#     sub_topics: List[str]
#     alias: str

# __slots__ = ('sub_topics',)

# ManFn = Callable[[Union[str, ...]], str]



import inspect
import sys
from abc import abstractmethod
from enum import Enum
from typing import TypeVar, Dict, Generator, Tuple, Any, Type, NoReturn

# from igit_debug.loggr import Loggr
from more_termcolor import colors

# from igit.abcs import prettyrepr
# from rich.console import Console
# from igit.util.misc import darkprint
from regexp import YES_OR_NO
import logging
T = TypeVar('T')
# from rich.console import Console
#
# console = Console()
# logger = Loggr(__name__)


# noinspection PyUnresolvedReferences
def mutate_identifier(identifier: str):
    upper = identifier.upper()
    identifier = upper
    logging.debug(f'mutate_identifier({repr(identifier)}) yielding upper: {repr(upper)}')
    yield upper
    words = self.val.split(' ')
    if len(words) == 1:
        raise NotImplementedError(f"no word separators, and both lowercase and uppercase identifier is taken ('{upper.lower()}')")
    words_identifiers = ''.join(map(lambda s: s[0], words))
    identifier = words_identifiers
    logging.debug(f'mutate_identifier() yielding words_identifiers: {repr(words_identifiers)}')
    yield words_identifiers
    for i in range(len(words_identifiers)):
        new_identifiers = words_identifiers[:i] + words_identifiers[i].upper() + words_identifiers[i + 1:]
        identifier = new_identifiers
        logging.debug(f'mutate_identifier() yielding new_identifiers (#{i}): {repr(new_identifiers)}')
        yield new_identifiers
    raise StopIteration(f'mutate_identifier() exhausted all options: {repr(self)}')


class Flow(Enum):
    """
    >>> flow = Flow('continue')
    <Flow.CONTINUE: 'continue'>
    >>> flow == Flow.CONTINUE == Flow(Flow.CONTINUE)
    True
    """
    CONTINUE = 'continue'
    DEBUG = 'debug'
    QUIT = 'quit'
    
    @staticmethod
    def laxinit(value) -> 'Flow':
        """Case insensitive"""
        if isinstance(value, str):
            return Flow(value.lower())
        return Flow(value)


# @prettyrepr
class Item:
    identifier: str
    value: Any
    
    def __init__(self) -> None:
        super().__init__()
        self._identifier = None
        self._value = None
    
    def __hash__(self) -> int:
        return hash((self.value, self.identifier))
    
    def __eq__(self, o: object) -> bool:
        if isinstance(o, (Item, Flow, str)):
            # both Item and Flow have a 'value' property
            return self.value == o
        
        return super().__eq__(o)
    
    def __str__(self):
        valuestr = str(self.value)
        if len(valuestr) >= 80:
            return valuestr[:77] + '...'
        return valuestr
    
    # def __repr__(self):  # uncomment if has prepr() fn (originally from @prettyrepr)
    #     identifier = f'[{self._identifier}]'
    #     return f"""{self.prepr()}({repr(str(self))}) | {identifier}"""
    
    @property
    def identifier(self) -> str:
        return self._identifier
    
    @identifier.setter
    def identifier(self, identifier: str):
        self._identifier = identifier
    
    @property
    def value(self) -> str:
        return self._value
    
    @value.setter
    def value(self, value: str):
        self._value = value


class MutableItem(Item):
    is_yes_or_no: bool
    
    def __init__(self, value, identifier=None) -> None:
        super().__init__()
        self.identifier = identifier
        self.value: str = value
    
    def __repr__(self):
        superrepr = repr(super())
        if bool(YES_OR_NO.fullmatch(self.identifier)):
            superrepr += f', is_yes_or_no: {True}'
        return superrepr
    
    def __hash__(self) -> int:
        return hash((self.value, self.identifier, self.is_yes_or_no))
    
    @property
    def is_yes_or_no(self):
        ret = bool(YES_OR_NO.fullmatch(self.identifier))
        # darkprint(f'{repr(self)}.is_yes_or_no() → {ret}')
        return ret
        # return self._is_yes_or_no
    
    # @property
    # def value(self):
    #     return self._value
    #
    # @value.setter
    # def value(self, value: str):
    #     # TODO: this is buggy, because:
    #     #  1. prompt.answer is bool if KEY.lower() == 'y', not val
    #     #  2. y='something' should register as yes or no, but doesn't
    #     self._value = value
    #     self._is_yes_or_no = bool(YES_OR_NO.fullmatch(value))


class LexicItem(MutableItem):
    def mutate_identifier(self):
        upper = self.identifier.upper()
        self.identifier = upper
        # darkprint(f'{repr(self)} | mutate_identifier() yielding upper: {repr(upper)}')
        yield upper
        words = self.value.split(' ')
        if len(words) == 1:
            raise NotImplementedError(f"no word separators, and both lowercase and uppercase identifier is taken ('{upper.lower()}')")
        words_identifiers = ''.join(map(lambda s: s[0], words))
        self.identifier = words_identifiers
        # darkprint(f'mutate_identifier() yielding words_identifiers: {repr(words_identifiers)}')
        yield words_identifiers
        for i in range(len(words_identifiers)):
            new_identifiers = words_identifiers[:i] + words_identifiers[i].upper() + words_identifiers[i + 1:]
            self.identifier = new_identifiers
            # darkprint(f'mutate_identifier() yielding new_identifiers (#{i}): {repr(new_identifiers)}')
            yield new_identifiers
        raise StopIteration(f'mutate_identifier() exhausted all options: {repr(self)}')


class FlowItem(Item):
    value: Flow
    
    def __init__(self, value) -> None:
        super().__init__()
        flow = Flow.laxinit(value)
        self._identifier = flow.value
        self._value = flow
    
    def __hash__(self) -> int:
        return hash((self.value, self.identifier))
    
    # def __repr__(self):  # uncomment if has prepr() fn (originally from @prettyrepr)
    #     # TODO: why is it not working with repr(super())?
    #     identifier = f'[{self._identifier}]'
    #     return f"""{self.prepr()}({repr(str(self))}) | {identifier}"""
    
    # def __init__(self, value):
    #     flow = Flow(value)
    #     self.identifier = flow.value
    #     self.value = flow.name
    
    # @classmethod
    # def full_names(cls) -> set:
    #     """→ {'continue', 'debug', 'quit'}"""
    #     return set(map(str.lower, FlowItem._member_names_))
    
    # @classmethod
    # def from_full_name(cls, fullname: str) -> 'FlowItem':
    #     """
    #     >>> FlowItem.from_full_name('continue')
    #     FlowItem.CONTINUE
    #     """
    #     try:
    #         return FlowItem._member_map_[str(fullname).upper()]
    #     except KeyError as e:
    #         raise ValueError(f"'{fullname}' is not a valid {cls.__qualname__}") from None
    
    @property
    def identifier(self):
        return self._identifier
    
    @identifier.setter
    def identifier(self, identifier):
        raise AttributeError(f"{repr(self)}.identifier({repr(identifier)}): Enum can't set self.identifier because self.value is readonly")
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value):
        raise AttributeError(f"{repr(self)}.value({repr(value)}): Enum can't set self.value because self.value is readonly")
    
    @property
    def DEBUG(self):
        return self.value is Flow.DEBUG
    
    @property
    def CONTINUE(self):
        return self.value is Flow.CONTINUE
    
    @property
    def QUIT(self):
        return self.value is Flow.QUIT
    
    def execute(self) -> None:
        print(f'execute() self: ', self)
        if self.QUIT:
            sys.exit('Aborting')
        if self.CONTINUE:
            print(colors.italic('continuing'))
            return None
        if self.DEBUG:
            frame = inspect.currentframe()
            up_frames = []
            while frame.f_code.co_filename == __file__:
                frame = frame.f_back
                up_frames.append(frame.f_code.co_name)
            print(colors.bold(f'u {len(up_frames)}'), colors.dark(repr(up_frames)))
            breakpoint()
            
            return None
        raise NotImplementedError(f"don't support this enum type yet: {self}")


class Items(Dict[str, MutableItem]):
    _itemcls = MutableItem
    
    def __new__(cls: Type['Items'], *args: Any, **kwargs: Any) -> 'Items':
        # TODO: apply items_gen here (scratch_MyDict.py), so can init and behave like normal dict (KeywordItems())
        # darkprint(f'{cls.__qualname__}.__new__(*args, **kwargs): {args}, {kwargs}')
        return super().__new__(cls, *args, **kwargs)
    
    def __init__(self, items):
        super().__init__()
        # logger.debug('items:', items)
        for value, identifier in self.items_gen(items):
            item = self._itemcls(value, identifier)
            self.store(item)
            # logger.debug(f'\titem: {item}')
    
    def __repr__(self):
        string = '{\n\t'
        for identifier, item in self.items():
            string += f'{repr(identifier)}: {repr(item)}\n\t'
        return string + '}'
    
    @abstractmethod
    def items_gen(self, items) -> Generator[Tuple[str, str], None, None]:
        ...
    
    def store(self, item: MutableItem):
        if item.identifier in self:
            try:
                self.mutate_until_unique(item)
            except AttributeError as e:
                # no item.mutate_identifier(), probably a FlowItem
                # mutate stored item instead
                stored_item = self.get(item.identifier)
                self[item.identifier] = item
                self.mutate_until_unique(stored_item)
                self[stored_item.identifier] = stored_item
            else:
                # item mutated successfully
                self[item.identifier] = item
        else:
            self[item.identifier] = item
    
    @abstractmethod
    def mutate_until_unique(self, item: MutableItem):
        ...
    
    def __setitem__(self, k: str, v) -> None:
        if k in self:
            existing_item = self.get(k)
            self[k] = v
            self.mutate_until_unique(existing_item)
        super().__setitem__(k, v)


class KeywordItems(Items):
    def items_gen(self, kw_items: dict):
        for kw, value in kw_items.items():
            # kw is identifier
            yield value, kw
    
    def mutate_until_unique(self, item: MutableItem):
        raise NotImplementedError(f"{repr(self)}.mutate_until_unique(item={repr(item)}) not implemented")


class NumItems(Items):
    def items_gen(self, num_items):
        for idx, value in enumerate(num_items):
            # idx is identifier
            yield value, str(idx)
    
    def mutate_until_unique(self, item: MutableItem) -> MutableItem:
        """Identifiers are sorted digits. Tries to fill a gap if exists, otherwise just appends to the list"""
        identifiers = sorted(map(int, self.keys()))
        for idx, identifier in enumerate(identifiers[:-1]):
            plus1 = identifier + 1
            if identifiers[idx + 1] > plus1:
                # there's a gap, e.g. [1, 2, 4, 5]
                item.identifier = str(plus1)
                break
        else:
            # reached end of loop with no holes
            item.identifier = str(identifiers[-1] + 1)
        return item


class LexicItems(Items):
    _itemcls = LexicItem
    
    def items_gen(self, lexic_items):
        for value in lexic_items:
            # initial is identifier
            yield value, value[0]
    
    def mutate_until_unique(self, item: LexicItem) -> NoReturn:
        try:
            for mutation in item.mutate_identifier():
                if mutation in self:
                    continue
                return
        except AttributeError as e:
            raise NotImplementedError
    # no item.mutate_identifier, probably a FlowItem



from abc import ABC
from typing import Union, NoReturn, Callable, Any, Iterable

from more_termcolor import colors

from .item import MutableItem, Items, Flow
from .item import NumItems, LexicItems, KeywordItems
from .util import has_duplicates
# from .util.misc import darkprint
from .item import FlowItem
from err import DevError


# @prettyrepr
class Options(ABC):
    items: Items
    _itemscls = None
    
    # flowopts: Tuple['FlowItem', ...]
    
    def __init__(self, *opts: str):
        if has_duplicates(opts):
            raise ValueError(f"{repr(self)} | __init__(opts) duplicate opts: ", opts)
        self.items = self._itemscls(opts)
        # self.flowopts = tuple()
        # self._values = None
        # self._items = None
        # self._indexeditems = None
        # self._all_yes_or_no = None  # calculated only when None (not True or False)
    
    def __bool__(self):
        # return bool(self.items) or bool(self.flowopts)
        return bool(self.items)
    
    # def __repr__(self): # uncomment if has prepr() fn (originally from @prettyrepr)
    #     return f'{self.prepr()}(items = {repr(self.items)})'
    
    def set_flow_opts(self, flowopts: Union[str, Iterable, bool]) -> NoReturn:
        """Sets `self.flowopts` with FlowItem objects.
        Handles passing different types of flowopts (single string, tuple of strings, or boolean), and returns a FlowItem tuple."""
        # darkprint(f'{self.__class__.__qualname__}.set_flow_opts(flowopts={repr(flowopts)})')
        if flowopts is True:
            flowitems = tuple(map(FlowItem, Flow.__members__.values()))
        elif isinstance(flowopts, str):  # 'quit'
            # flowopts = (FlowItem.from_full_name(flowopts),)
            flowitems = (FlowItem(flowopts),)
        else:
            # flowopts = tuple(map(FlowItem.from_full_name, flowopts))
            flowitems = tuple(map(FlowItem, flowopts))
            if has_duplicates(flowitems):
                raise ValueError(f"{repr(self)}\nset_flow_opts(flowopts) | duplicate flowitems: {repr(flowitems)}")
        
        for flowitem in flowitems:
            self.items.store(flowitem)
            # if flowopt.value in self.items:
            #     raise ValueError(f'{repr(self)}\nset_special_options() | flowopt.value ({repr(flowopt.value)}) already exists in self.\nflowopt: {repr(flowopt)}.\nflowopts: {repr(flowopts)}')
            # self.items[flowopt.value] = flowopt.name
        return None
    
    def set_kw_options(self, **kw_opts: Union[str, tuple, bool]) -> None:
        """foo='bar', baz='continue'"""
        # darkprint(f'{self.__class__.__qualname__}.set_kw_options(kw_opts={repr(kw_opts)})')
        if not kw_opts:
            return
        if has_duplicates(kw_opts.values()):
            raise ValueError(f"{repr(self)}\nset_kw_options() duplicate kw_opts: {repr(kw_opts)}")
        if 'free_input' in kw_opts:
            raise DevError(f"{repr(self)}\nset_kw_options() | 'free_input' found in kw_opts, should have popped it out earlier.\nkw_opts: {repr(kw_opts)}")
        non_flow_kw_opts = dict()
        for kw in kw_opts:
            opt = kw_opts[kw]
            if kw in self.items:
                raise ValueError(f"{repr(self)}\nset_kw_options() | '{kw}' in kw_opts but was already in self.items.\nkw_opts: {repr(kw_opts)}")
            
            try:
                # flowitem = FlowItem.from_full_name(opt)
                flowitem = FlowItem(opt)
            except ValueError:
                non_flow_kw_opts[kw] = opt
            else:
                # not using self.items.store(flowitem) because kw isn't flowitem.identifier
                self.items[kw] = flowitem
        
        if not non_flow_kw_opts:
            return
        
        kw_items = KeywordItems(non_flow_kw_opts)
        
        self.items.update(**kw_items)
        # self.kw_opts = kw_opts
    
    def any_item(self, predicate: Callable[[MutableItem], Any]) -> bool:
        for item in self.items.values():
            if predicate(item):
                return True
        return False
    
    def all_yes_or_no(self) -> bool:
        
        for item in self.items.values():
            try:
                if not item.is_yes_or_no:
                    return False
            except AttributeError:
                print(colors.brightblack(f'{self.__class__.__qualname__}.all_yes_or_no() AttributeError with item.is_yes_or_no: {item} {type(item)}. Ignoring.'))
        return True
        
        # nonspecials = set(self.opts)
        # nonspecials.update(set(self.kw_opts.values()))
        # if not nonspecials:
        #     return False
        # all_yes_or_no = True
        # for nonspecial in nonspecials:
        #     if not re.fullmatch(YES_OR_NO, nonspecial):
        #         all_yes_or_no = False
        #         break
        # return all_yes_or_no
    
    # @cachedprop
    # def indexeditems(self) -> dict:
    #     # TODO: create an Items class (maybe also an MutableItem class?)
    #     indexeditems = dict()
    #     for idx, opt in enumerate(self.opts):
    #         indexeditems[str(idx)] = opt
    #
    #         # * self.kw_opts
    #     self._update_kw_opts_into_items(indexeditems)
    #
    #     # * self.flowopts
    #     self._update_special_opts_into_items(indexeditems)
    #
    #     return indexeditems
    
    # @cachedprop
    # def items(self) -> dict:
    #     # assumes duplicates between opts, special and kw were checked already
    #     items = dict()
    #     # * self.opts
    #     initials = [o[0] for o in self.opts]  # initials: 'w' 'w' 's' 'i' 'a'
    #     for idx, opt in enumerate(self.opts):
    #         initial: str = opt[0]
    #         duplicate_idxs = [jdx for jdx, jinitial in enumerate(initials) if jinitial == initial and jdx != idx]
    #         if not duplicate_idxs:
    #             items[initial] = opt
    #             continue
    #
    #         if len(duplicate_idxs) == 1:
    #             if initial.isupper():
    #                 # should handle like >= 2 duplicates
    #                 raise NotImplementedError("duplicate uppercase, probably one was uppercased in prev iteration: ", self.opts)
    #             # * opt = 'w', duplicates = ['w']
    #             #    transform one to 'W'
    #             upper = initial.upper()
    #             initials[idx] = upper
    #             items[upper] = opt
    #             dup_idx = duplicate_idxs[0]
    #             duplicate = initials[dup_idx]
    #             items[duplicate] = self.opts[dup_idx]
    #             continue
    #
    #         if len(duplicate_idxs) == 2:
    #             words = opt.split(' ')
    #             joined = ''.join(map(lambda s: s[0], words))
    #             if joined in initials:
    #                 for j in range(len(words)):
    #                     new_joined = joined[:j] + joined[j].upper() + joined[j:]
    #                     if new_joined not in initials:
    #                         joined = new_joined
    #                         break
    #                 else:
    #                     raise NotImplementedError("duplicate multi words with no unique uppercase permutation: ", self.opts)
    #
    #             items[joined] = opt
    #             dup_idx1, dup_idx2 = duplicate_idxs
    #             upper = initials[dup_idx1].upper()
    #             initials[dup_idx1] = upper
    #             items[upper] = self.opts[dup_idx1]
    #             items[initials[dup_idx2]] = self.opts[dup_idx2]
    #             continue
    #
    #         raise NotImplementedError("3 duplicate options with no word separators: ", self.opts)
    #
    #     # * self.kw_opts
    #     self._update_kw_opts_into_items(items)
    #
    #     # * self.flowopts
    #     self._update_special_opts_into_items(items)
    #
    #     return items
    
    # def _update_kw_opts_into_items(self, items: dict):
    #     for k, opt in self.kw_opts.items():
    #         if k in items:
    #             raise NotImplementedError("kw_opts clashes with items. displaying self.opts, self.kw_opts and items:",
    #                                       self.opts, self.kw_opts, items)
    #         items[k] = opt
    #
    # def _update_special_opts_into_items(self, items: dict):
    #     for spec in self.flowopts:
    #         if spec.value in items:
    #             raise NotImplementedError("flowopts clashes with items. displaying self.opts, self.kw_opts, flowopts and items:",
    #                                       self.opts, self.kw_opts, self.flowopts, items)
    #         items[spec.value] = spec.name


class NumOptions(Options):
    _itemscls = NumItems
    items: NumItems
    
    # def __init__(self, *opts: str):
    #     self.items = NumItems(opts)
    #     super().__init__(*opts)


class LexicOptions(Options):
    _itemscls = LexicItems
    items: LexicItems
    
    # def __init__(self, *opts: str):
    #     self.items = LexicItems(opts)
    #     super().__init__(*opts)



#!/usr/bin/env python3.8
"""
get_main_topics() is called to popuplate MAIN_TOPICS.
"""
import inspect
import logging
import re
import sys
# from functools import wraps
from typing import Callable, Union
from typing import Dict, List, Any, Tuple

import click
# from more_termcolor import cprint

# import search
# import prompt
from manuals.common.click_extension import unrequired_opt

# stuff from manuals.py to ignore when dir'ing manuals module
from manuals.common.types import ManFn


# brightprint = lambda s, *colors: cprint(s, 'bright white', *colors)
SUB_TOPIC_RE = re.compile(r'_[A-Z]*\s?=\s?(rf|fr|f)"""')
from rich.console import Console

console = Console()
EXCLUDE = [
    # imports
    'inspect',
    're',
    
    # functools
    'wraps',
    
    # typing
    'Literal',
    'Dict',
    'Type',
    'Union',
    
    # formatting
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'b',
    'c',
    'i',
    'black',
    'bg',
    'ManFn',
    
    # pygments
    'Lexer',
    'pyglight',
    'TerminalTrueColorFormatter',
    'get_lexer_by_name',
    'BashLexer',
    'CssLexer',
    'DockerLexer',
    'IniLexer',
    'JavascriptLexer',
    'JsonLexer',
    'MySqlLexer',
    'PythonLexer',
    'SassLexer',
    'TypeScriptLexer',
    
    # module-level vars
    'linebreak',
    'backslash',
    # 'color',
    'Style',
    'Language',
    'langs',
    'HIGHLIGHT_START_RE',
    'HIGHLIGHT_END_RE',
    'formatters',
    'lexers',
    
    # functions
    '_get_lexer_ctor',
    '_get_lexer',
    '_get_color_formatter',
    '_highlight',
    'alias',
    'syntax',
    
    # module magics
    '__annotations__',
    '__builtins__',
    '__cached__',
    '__doc__',
    '__file__',
    '__loader__',
    '__name__',
    '__package__',
    '__spec__'
    ]
# perform_checks = any(arg == '--check' for arg in sys.argv)


# module = dict(igit=dict())


# # noinspection PyUnresolvedReferences
# # This is only called by itself right now, no actual usage
# def fromigit(dotnames, igitmod=None, module_subdict: dict = None):
#     """
#     Examples
#     ::
#         fromigit('prompt')
#         fromigit('prompt.Flow')
#         fromigit('util.clickex.unrequired_opt')
#     """
#     # TODO: IMPLEMENT
#     #  the goal is to avoid actual import as much as possible
#     #  so that `module` stores only dicts if not used directly, e.g. 'util' can be a { 'clickex': {'unrequired_opt': unrequired_opt} } where
#     #  only unrequired_opt is a function.
#     #  it's tricky because 'prompt' can be used directly e.g. prompt.choose, or indirectly e.g. prompt.Flow.CONTINUE
#     #  need to decide, if I allow fromigit('prompt').choose then fromigit('prompt') must be a module and not a dict.
#     #  in which case, how to avoid importing too much when fromigit('prompt.Flow')?
#
#     ## possible scenarios:
#     # dotnames = 'prompt', igitmod = None, module_subdict = None →
#     #   dotnames = 'prompt', igitmod = igit, module_subdict = module['igit'] →
#
#     # dotnames = 'prompt.Flow', igitmod = None, module_subdict = None →
#     #   dotnames = 'prompt.Flow', igitmod = igit, module_subdict = module['igit'] →
#
#     # dotnames = 'util.clickex.unrequired_opt', igitmod = None, module_subdict = None →
#
#     if module_subdict is None:
#         # assuming igitmod is None also
#         import igit
#         return fromigit(dotnames, igitmod=igit, module_subdict=module['igit'])
#
#     name, _, nestednames = dotnames.partition('.')
#     cachedval = module_subdict.get(name)
#     if cachedval is None:
#         nestedmod = getattr(igitmod, name)  # actual import, so nestedmod may be prompt
#         module_subdict[name] = dict()  # initialize e.g. module['igit']['prompt'] = prompt
#
#         if nestednames:
#             # e.g. nestednames='clickex.unrequired_opt'; nestedmod=util; module_subdict[name]=dict()
#             val = fromigit(nestednames, igitmod=nestedmod, module_subdict=module_subdict[name])
#         else:
#             # this is it
#             val = nestedmod
#             module_subdict[name] = val
#             return val
#
#     else:
#         ## cachedval is a dict
#         if nestednames:
#             # e.g. nestednames = 'Flow', cachedval = module['igit']['prompt']
#             val = fromigit(nestednames, igitmod=nestedmod, module_subdict=module_subdict[name])
#         else:
#             # e.g. cachedval = module['igit']['prompt']
#             return cachedval
#         val = nestedmod
#     if first_recursion:
#         module[nestednames] = val
#     if not val:
#         errmsg = (f"fromigit(dotnames: '{dotnames}', igitmod: {igitmod}) | "
#                   "locals:"
#                   f"{locals()}"
#                   )
#         raise ModuleNotFoundError(errmsg)
#     return val


# def logger() -> ForwardRef('Loggr'):
#     _logger = module.get('logger')
#     if _logger is not None:
#         return _logger
#     # 'logger' not in module; construct one
#     IGIT_LOG_LEVEL = os.environ.get('IGIT_LOG_LEVEL', '')
#     if IGIT_LOG_LEVEL.lower() == 'none':
#         class Loggr:
#             def __init__(self, *args, **kwargs):
#                 pass
#
#             def __getattribute__(self, item):
#                 return self
#
#             def __call__(self, *args, **kwargs):
#                 return self
#
#     else:
#         print('myman.py importing igit_debug Loggr')
#         from igit_debug.loggr import Loggr
#     module['logger'] = Loggr(__name__)
#     return module['logger']


# used by print_manual()
# def printrv(fn):
#     @wraps(fn)
#     def wrap(*args, **kwargs):
#         rv = fn(*args, **kwargs)
#         print(rv)
#         return rv
#
#     return wrap


def create_new_manual(newtopic: str):  # called in __main__.py
    logging.debug(f'create_new_manual({newtopic = })')
    if newtopic.lower() in map(str.lower, MAIN_TOPICS):
        logging.warning(f"{newtopic = } already exists")
        return False
    logging.warning('should not be implemented, dont be lazy')
    # from mytool.myman import manuals
    # print(f"creating {newtopic} in {manuals.__file__}...")
    # with open(manuals.__file__, mode='r+') as manfile:

def get_skipped_subtopics(undecorated_main_topic_fn):
    """python3.8 -m mytool.myman --check calls this"""
    lines = inspect.getsource(undecorated_main_topic_fn).splitlines()
    last_else_idx = -next(i for i, line in enumerate(reversed(lines)) if line.strip() == 'else:') - 1
    before_else = [line.strip().partition('=')[0].strip() for line in lines[:last_else_idx] if
                   line and SUB_TOPIC_RE.search(line)]
    after_else = [line.strip()[1:-1] for line in lines[last_else_idx:] if
                  (stripped := line.strip()).startswith('{_') and stripped.endswith('}')]
    if diff := set(before_else).difference(set(after_else)):
        logging.warning(f'{undecorated_main_topic_fn.__name__}() doesnt print: {diff}')


def draw_out_decorated_fn(fn: ManFn) -> Callable:
    closure: tuple = fn.__closure__
    if not closure:
        # non-decorated functions' closure is None
        return fn
    return closure[-1].cell_contents


def populate_main_topics() -> Dict[str, ManFn]:
    """Populates a { 'pandas' : pandas , 'inspect' : inspect_, 'gh' : githubcli } dict from `manuals` module"""
    from . import manuals
    main_topics = dict()
    for main_topic in dir(manuals):
        if main_topic in EXCLUDE:
            continue
        fn: Callable = getattr(manuals, main_topic)
        if not inspect.isfunction(fn):
            continue
        # noinspection PyUnresolvedReferences
        if not fn.__module__.endswith('.manuals') or 'Lexer' in main_topic:
            # This happens when something is missing in EXCLUDE, oversight from my part
            continue
        if main_topic.endswith('_'):  # inspect_()
            name = main_topic[:-1]
        else:
            name = main_topic
        # * dont draw_out_decorated_fn because then syntax() isn't called
        main_topics[name] = fn
        try:
            alias = getattr(fn, 'alias')
        except:
            pass
        else:
            main_topics[alias] = fn
    
    return main_topics


MAIN_TOPICS: Dict[str, ManFn] = populate_main_topics()


def get_sub_topic_var_names(fn: ManFn) -> List[str]:
    return [n for n in fn.__code__.co_varnames if n.isupper()]


def populate_sub_topics(*, print_skipped_subtopics=False) -> Dict[str, Union[List[ManFn], ManFn]]:
    """Sets bash.sub_topics = [ "cut" , "for" ] for each MAIN_TOPICS.
    Removes (d)underscore and lowers _CUT and __FOR.
    Returns e.g. `{ 'cut' : bash , 'args' : [ bash , pdb ] }`"""
    # var_reg = r'_[A-Z0-9_]*'
    # matches `_SUBTOPIC = fr"""..."""` and `_ST = _SUBTOPIC`
    # sub_topic_reg = re.compile(fr'\s*{var_reg}\s?=\s?(r?fr?""".*|{var_reg})\n')
    # subject_sig_re = re.compile(r'subject\s?=\s?None')
    
    sub_topics = dict()
    # populate sub topics as main_topics
    # global perform_checks
    for main_topic_fn in list(MAIN_TOPICS.values()):
        
        # sub_topic_var_names = [n for n in draw_out_decorated_fn(main_topic_fn).__code__.co_varnames if n.isupper()]
        # draw_out_decorated_fn is necessary because MAIN_TOPICS has to store the fns in the wrapped form (to call them later)
        setattr(main_topic_fn, 'sub_topics', [])
        undecorated_main_topic_fn = draw_out_decorated_fn(main_topic_fn)
        sub_topic_var_names = get_sub_topic_var_names(undecorated_main_topic_fn)
        if not sub_topic_var_names:
            continue
        # lines, _ = inspect.getsourcelines(main_topic_fn)
        # if not re.search(subject_sig_re, lines[0]) and not re.search(subject_sig_re, lines[1]):
        # lines[1]: decorated functions' first line is the @decorator
        # if not 'subject' in inspect.getfullargspec(main_topic_fn).args:
        #     continue
        
        # lines = [l.strip() for l in lines if re.fullmatch(sub_topic_reg, l)]
        # if not lines:  # has subject=None in sig but no actual sub topic vars
        #     continue
        
        if print_skipped_subtopics:
            print(get_skipped_subtopics(undecorated_main_topic_fn))
        
        for sub_topic_var_name in sub_topic_var_names:
            # DONT account for dunder __ARGUMENTS, it's being handled by get_sub_topic().
            # If handled here, eventually bash('_ARGUMENTS') is called which errors.
            # if (stripped := sub_topic_var_name.strip()).startswith('__'):
            #     sub_topic_var_name = stripped.lower()[2:]
            # else:  # starts with single '_'
            #     sub_topic_var_name = stripped.lower()[1:]
            sub_topic_var_name = sub_topic_var_name.strip().lower()[1:]
            main_topic_fn.sub_topics.append(sub_topic_var_name)
            if sub_topic_var_name in sub_topics:
                # this means duplicate subtopic, for different main topics
                # in that case, the value is set to be a list of subtopics
                if isinstance(sub_topics[sub_topic_var_name], list):
                    sub_topics[sub_topic_var_name].append(main_topic_fn)
                else:
                    # create a new list
                    sub_topics[sub_topic_var_name] = [sub_topics[sub_topic_var_name], main_topic_fn]
            else:
                # no duplicate sub topics; value is set to be simply the function
                sub_topics[sub_topic_var_name] = main_topic_fn
    
    return sub_topics


SUB_TOPICS: Dict[str, Union[List[ManFn], ManFn]] = populate_sub_topics()


def fuzzy_find_topic(topic, collection, *extra_opts, raise_if_exhausted=False, **extra_kw_opts) -> Tuple:
    """If user continue'd through the whole collection, raises KeyError if `raise_if_exhausted` is True. Otherwise, returns None"""
    # not even a subtopic, could be gibberish
    # try assuming it's a substring
    import search
    import prompt
    for maybes, is_last in search.iter_maybes(topic, collection, criterion='substring'):
        if not maybes:
            continue
        
        kwargs = dict(Ep='Edit manuals.py with pycharm', Ec='Edit manuals.py with vscode',
                      Em='Edit manuals.py with micro')
        
        if is_last and raise_if_exhausted:
            kwargs.update(dict(flowopts='quit'))
        else:
            kwargs.update(dict(flowopts='quit', no='continue'))
        # TODO: if a match is a subtopic, present maintopic.subtopic in choose
        key, choice = prompt.choose(f"Did you mean any of these?", *extra_opts, *maybes, **kwargs,
                                    **extra_kw_opts)
        if choice == prompt.Flow.CONTINUE:
            continue
        if isinstance(key, str) and key.startswith('E'):
            import os
            manualsfile = os.path.join(os.path.dirname(__file__), 'manuals.py')
            if key == 'Ep':
                os.system(f'pycharm "{manualsfile}"')
            if key == 'Ec':
                os.system(f'code "{manualsfile}"')
            if key == 'Em':
                os.system(f'micro "{manualsfile}"')
            sys.exit()
        return key, choice.value
    if raise_if_exhausted:
        raise KeyError(f"'{topic}' isn't in collection")
    return None, None


def get_sub_topic(main_topic: str, sub_topic: str):
    """
    MAIN EXISTS?
     |      \
    yes     no
     |       \
     |      [find main]
     |        |     \
     |      found   nope
     |       /        \
     |     /        KeyError!
     |   /
    SUB IN MAIN's SUBS?
     |          \
    yes         no
     |           \
    [return]    [find sub in main's]
                   |         \
                found       nope
                  |           \
                [return]    SUB IN ANY MAIN's SUBS?
                              |         \
                            yes         no
                             |           \
                            [return]    [find sub]
                                          |       \
                                        found      nope
                                          |         \
                                        [return]   KeyError!
    """
    logging.debug(f"get_sub_topic({repr(main_topic)}, {repr(sub_topic)})")
    if main_topic not in MAIN_TOPICS:
        console.log(f"[bright_white]Unknown main topic: '{main_topic}'. Searching among MAIN_TOPICS...[/]")
        # brightprint(f"Unknown main topic: '{main_topic}'. Searching among MAIN_TOPICS...")
        main_topic = fuzzy_find_topic(main_topic, MAIN_TOPICS, raise_if_exhausted=True)
    
    main_topic_fn: ManFn = MAIN_TOPICS[main_topic]
    if sub_topic in main_topic_fn.sub_topics:
        return main_topic_fn(f'_{sub_topic.upper()}')
    if f'_{sub_topic}' in main_topic_fn.sub_topics:
        # mm bash syntax → main_topic_fn.sub_topics has '_syntax' → pass '__SYNTAX'
        return main_topic_fn(f'__{sub_topic.upper()}')
    
    try:
        # sometimes functions have alias logic under 'if subject:' clause, for example bash
        # has 'subject.startswith('<')'. so maybe sub_topic works
        return main_topic_fn(sub_topic)
    except KeyError:
        console.log((f"[bright_white]sub topic '{sub_topic}' isn't a sub topic of '{main_topic}'. "
                     f"Searching among '{main_topic}'s sub topics...[/]"))
        key, chosen_sub_topic = fuzzy_find_topic(sub_topic,
                                                 main_topic_fn.sub_topics,
                                                 raise_if_exhausted=False,
                                                 # keep uppercase P so doesn't collide with topics
                                                 P=f"print '{main_topic}' w/o subtopic")
        if key == 'P':
            return main_topic_fn()
        
        if chosen_sub_topic is not None:
            return main_topic_fn(f'_{chosen_sub_topic.upper()}')
        
        if sub_topic in SUB_TOPICS:
            console.log(f"[bright_white]'{sub_topic}' isn't a sub topic of '{main_topic}', but it belongs to these topics:[/]")
            return print_manual(sub_topic)
        
        console.log(f"[bright_white]'{sub_topic}' doesn't belong to any topic. Searching among all SUB_TOPICS...[/]")
        key, sub_topic = fuzzy_find_topic(sub_topic,
                                          SUB_TOPICS,
                                          raise_if_exhausted=True,
                                          P=f"print '{main_topic}' w/o subtopic")
        if key == 'P':
            return main_topic_fn()
        
        return main_topic_fn(f'_{sub_topic.upper()}')


def print_manual(main_topic: str, sub_topic=None):
    """If passed correct topic(s), prints.
    If not correct, finds the correct with fuzzy search and calls itself."""
    logging.debug(f"print_manual({repr(main_topic)}, {repr(sub_topic)})")
    if sub_topic:
        # * passed both main, sub
        sub_topic_str = get_sub_topic(main_topic, sub_topic)
        # sub_topic_str = get_sub_topic(main_topic, sub_topic) \
        #     .replace('[h1]', '[bold underline reverse bright_white]') \
        #     .replace('[h2]', '[bold underline bright_white]') \
        #     .replace('[h3]', '[bold bright_white]') \
        #     .replace('[h4]', '[bright_white]') \
        #     .replace('[h5]', '[white]') \
        #     .replace('[c]', '[dim]')
        # return console.print(sub_topic_str)
        return print(sub_topic_str)
    # ** passed only one arg, could be main or sub
    try:
        # * assume precise main topic, i.e. "git"
        return print(MAIN_TOPICS[main_topic]())
    except KeyError as not_main_topic:
        # * so maybe it's a precise sub topic, i.e. "diff"
        topic = main_topic
        try:
            fn: Callable[[str], Any] = SUB_TOPICS[topic]
        except KeyError as not_sub_topic_not_main_topic:
            # * not a precise subtopic. find something precise, either a main or sub topic
            key, topic = fuzzy_find_topic(topic,
                                          list(MAIN_TOPICS.keys()) + list(SUB_TOPICS.keys()),
                                          raise_if_exhausted=True)
            return print_manual(topic)
        else:
            # * indeed a precise subtopic, call the main topic's fn and pass subtopic param
            try:
                return print(fn(f'_{topic.upper()}'))
            except TypeError as duplicate_sub_topic:
                # * error: "'list' object is not callable". some topics share same subtopic. choose main topic
                import prompt
                fn: List[Callable[[str], Any]]
                
                # TODO (bugs): 
                #  (1) If an ALIAS of a subtopic is the same as a SUBTOPIC of another main topic,
                #    this is called (shouldn't). Aliases aren't subtopics. (uncomment asyncio # _SUBPROCESS = _SUBPROCESSES)
                #  (2) main topics with both @alias and @syntax decors, that have the issue above ("(1)"), raise
                #    a ValueError in igit prompt, because the same main topic function is passed here for each subtopic and subtopic alias.
                #  ValueError: ('NumOptions | __init__(opts) duplicate opts: ', ('<function asyncio at 0x7f5dab684ca0>', '<function asyncio at 0x7f5dab684ca0>', '<function python at 0x7f5dab669a60>'))
                idx, choice = prompt.choose(f"'{topic}' exists in several topics, which one did you mean?",
                                            *[f.__qualname__ for f in fn],
                                            flowopts='quit'
                                            )
                return print(fn[idx](f'_{topic.upper()}'))





@click.command(context_settings=dict(help_option_names=['-h', '--help']), help="--check is also possible")
@click.argument('main_topic')
@click.argument('sub_topic', required=False)
@unrequired_opt('-l', '--list', 'list_subtopics', is_flag=True, help='list sub topics')
@unrequired_opt('--doctor', 'print_skipped_subtopics', is_flag=True, help="prints any subtopics that are skipped erroneously in a main topic's else clause")
def get_topic(main_topic, sub_topic, list_subtopics, print_skipped_subtopics):
    logging.debug(f'myman.get_topic({main_topic = }, {sub_topic = }, {list_subtopics = })')
    if print_skipped_subtopics:
        populate_sub_topics(print_skipped_subtopics=True)
        return
    if list_subtopics:
        console.log(f"[bright_white bold underline]{main_topic}[/]\n")
        [print(f'{st}') for st in sorted(MAIN_TOPICS[main_topic].sub_topics)]
        return
    
    print_manual(main_topic, sub_topic)



from collections import defaultdict
from typing import Literal, Collection, TypeVar, Optional, List, Callable, Tuple, Generator, Generic, Dict
import logging
import re
from fuzzysearch import find_near_matches
import regexp

T = TypeVar('T')

SearchCriteria =  Literal['substring', 'equals', 'startswith', 'endswith']


class Matches(Generic[T]):
    def __init__(self, *, maxsize):
        self.count = 0
        self.matches: Dict[float, List[T]] = defaultdict(list)
        self.worst_score = 0
        self.best_score = 999
        self._maxsize = maxsize
    
    def __bool__(self):
        return bool(self.matches)
    
    def __repr__(self):
        matches_repr = ''
        for score, candidates in self.matches.items():
            matches_repr += f'[{score}]: {", ".join(candidates)}\n    '
        return f"""Matches() ({self.count}) | best: {self.best_score} | worst: {self.worst_score}
    {matches_repr}"""
    
    def _reset_stats(self):
        # don't update best_score because we're called after discarding only worst scores
        # print(paint.faint(f'Matches._reset_stats() beforehand: worst {self.worst_score}, count {self.count}'))
        
        self.worst_score = 0
        self.count = 0
        for score, candidates in self.matches.items():
            if score > self.worst_score:
                self.worst_score = score
            self.count += len(candidates)
    
    def append(self, item: T, score: float):
        # lower score is better
        if self.count < self._maxsize:
            # * below maxsize
            if score > self.worst_score:
                self.worst_score = score
            if score < self.best_score:
                self.best_score = score
            self.matches[score].append(item)
            self.count += 1
            return
        
        # * reached maxsize
        if score == self.worst_score:
            # even if reached max size, append item to [worst_score] because no way to decide what to discard
            self.matches[self.worst_score].append(item)
            self.count += 1
            return
        
        if score < self.worst_score:
            # better than worst; discard worst candidates, add current item
            del self.matches[self.worst_score]
            self.matches[score].append(item)
            if score < self.best_score:
                self.best_score = score
            self._reset_stats()
        
        # score is worst than worst: don't let in
    
    def best(self) -> Optional[List[T]]:
        if not self.matches:
            logging.debug(f'no self.matches → returning None')
            return None
        ret = self.matches[self.best_score]
        # print(paint.faint(f'Matches.best() → {len(ret)} matches with score: {self.best_score}'))
        return ret
def _create_is_maybe_predicate(criterion: SearchCriteria) -> Callable[[str, str], bool]:
    if criterion == 'substring':
        is_maybe = str.__contains__
    elif criterion == 'equals':
        is_maybe = str.__eq__
    elif criterion == 'startswith':
        is_maybe = str.startswith
    elif criterion == 'endswith':
        is_maybe = str.endswith
    else:
        raise ValueError(f'Expected criterion: {SearchCriteria.__args__}, got "{criterion}"')
    
    return is_maybe


def fuzzy(keyword: str, collection: Collection[T], cutoff=2) -> Matches[T]:
    """Returns a `Matches` instance, or None if nothing matched.
    Doesn't prompt.
    """
    if not keyword or re.fullmatch(r'[\'"]+', keyword):  # only quotes
        raise ValueError(f"fuzzy(keyword={repr(keyword)}): bad keyword")
    if not collection or not any(item for item in collection):
        raise ValueError(f"fuzzy({repr(keyword)}, collection = {repr(collection)}): no collection")
    near_matches = Matches(maxsize=5)
    far_matches = Matches(maxsize=5)
    max_l_dist = min(len(keyword) - 1, 17)
    # TODO: sometimes cutoff == max_l_dist
    coll_len = len(collection)
    logging.debug(f'fuzzy({repr(keyword)}, collection ({coll_len}), cutoff: {cutoff}, max_l_dist: {max_l_dist})')
    printed_progress = -1
    for i, item in enumerate(collection):
        progress = round(i / coll_len, 1)
        if progress > printed_progress:
            logging.debug(f'{progress * 100}% (near_matches: {near_matches.count}, far_matches: {far_matches.count})')
            if near_matches.count ^ far_matches.count:
                # exactly one is Truthy
                coll = next(filter(bool, [near_matches, far_matches]))
                logging.debug(repr(coll))
            printed_progress = progress
        matches = find_near_matches(keyword, item, max_l_dist=max_l_dist)
        # don't `continue` even if matches is empty
        # lower distance is better
        dists_sum = 0
        # dists = []
        match_lengths = 0
        match_count = 0
        for m in matches:
            dists_sum += m.dist
            # dists.append(m.dist)
            match_lengths += len(m.matched)
            match_count += 1
        if not dists_sum:
            continue
        
        # dist_score = sum(dists) / len(dists)
        dist_score = dists_sum / match_count
        avg_match_length = match_lengths / match_count
        relative_factor = avg_match_length / len(item)
        relative_factor = -round(-relative_factor - (-relative_factor % 0.2), 2)  # round up
        # relative_factor = 1 + (dist_score / len(item))
        # score = dist_score * relative_factor
        
        score = dist_score - relative_factor
        if score >= cutoff:
            # * not so good (above cutoff): put into far_matches if within cutoff+1
            far_matches.append(item, score)
            continue
        
        # * good (below cutoff)
        near_matches.append(item, score)
    if not near_matches and not far_matches:
        logging.warning(f'GOT NOTHING! no near_matches, no far_matches. returning empty Matches object. collection:', collection)
        return near_matches
    if near_matches:
        logging.debug(f'fuzzy() → near_matches')
        return near_matches
    logging.debug(f'fuzzy() → far_matches')
    return far_matches
def iter_maybes(keyword: str, collection: Collection[T], *extra_options, criterion: SearchCriteria = 'substring') -> Generator[Tuple[List[T], bool], None, None]:
    """Doesn't prompt. Yields three `[matches...], is_last` tuples.
    1st: str method by `criterion`
    2nd: re.search ignoring word separators
    3rd: fuzzy search (what `nearest()` uses directly)"""
    is_maybe = _create_is_maybe_predicate(criterion)
    
    logging.debug(f'iter_maybes({repr(keyword)}, collection ({len(collection)}), extra_options: ({len(extra_options)}), criterion: {repr(criterion)}), is_maybe: {is_maybe.__name__}')
    if extra_options:
        logging.warning(f'got extra_options, ignored because not developed:', extra_options)
    maybes = [item for item in collection if is_maybe(item, keyword)]
    logging.debug(f'yielding {len(maybes)} maybes that True for `is_maybe(item, keyword)`')
    yield maybes, False
    
    regexp = None
    try:
        regexp_str = re.sub(r'[-_./ ]', r'[-_./ ]?', keyword)
        regexp = re.compile(regexp_str, re.IGNORECASE)
    except re.error:
        # compilation failed, probaby keyword is itself a regexp
        if regexp.has_regex(keyword):
            regexp = re.compile(keyword, re.IGNORECASE)
    
    if regexp is not None:
        new_maybes = []
        for item in collection:
            if regexp.search(item) and item not in maybes:
                new_maybes.append(item)
        if new_maybes and new_maybes != maybes:
            logging.debug(f"regexp isn't None ({regexp}) and searching resulted in different maybes. yielding {len(new_maybes)} new maybes")
            yield new_maybes, False
    
    near_matches = fuzzy(keyword, collection)
    if near_matches:
        logging.debug(f"last stop: yielding `(near_matches.best(), True)`. near_matches: {repr(near_matches)}")
        yield near_matches.best(), True
    else:
        logging.warning(f"near_matches is None for fuzzy({repr(keyword)}, collection ({len(collection)})). Will throw AttributeError")
        yield None, True



import click
import typing


# import functools as ft


def option(*param_decls, **attrs):
    """`show_default = True`.

    If `default` in `attrs` and `type` is not, or vice versa, sets one based on the other.

    Unless `default = None`, in which case `type` isn't set.

    If both are missing, `type` defaults to `str`.


    `type` can be either:
     - type, i.e. `str`
     - tuple, i.e. `(str, int)`
     - `typing.Literal['foo']`
     - `click.typing.<Foo>` (which includes click.Choice(...))
    """
    attrs['show_default'] = True
    default = attrs.get('default', click.core._missing)
    default_is_missing = default is click.core._missing
    typeattr = attrs.get('type', click.core._missing)
    type_is_missing = typeattr is click.core._missing
    if not type_is_missing:
        if typing.get_origin(typeattr) is typing.Literal:
            # type=typing.Literal['foo']. build a click.Choice from it
            attrs['type'] = click.Choice(typeattr.__args__)
            if default_is_missing:
                # take first Literal arg
                attrs['default'] = typeattr.__args__[0]
        
        else:
            # not a typing.Literal (e.g. `type=str`)
            attrs['type'] = typeattr
            if default_is_missing:
                attrs['default'] = typeattr()
    
    else:
        # type is missing.
        # if default=None, it's probably just a placeholder and
        # doesn't tell us above the 'real' type
        if not default_is_missing and default is not None:
            attrs['type'] = type(default)
        # otherwise, type and default both missing. not sure if this works
    
    if attrs.get('metavar', click.core._missing) is click.core._missing \
            and attrs.get('type', click.core._missing) is not click.core._missing:
        try:
            # changes click's default "BOOLEAN" to "BOOL", "INTEGER" → "INT"
            attrs['metavar'] = attrs['type'].__name__.upper()
        except AttributeError:
            # has no attribute __name__
            pass
    return click.option(*param_decls, **attrs)


def unrequired_opt(*param_decls, **attrs):
    """`required = False, show_default = True`.

    If `default` in `attrs` and `type` is not, or vice versa, sets one based on the other.

    Unless `default = None`, in which case `type` isn't set.

    If both are missing, `type` defaults to `str`.

    `type` can be either:
     - type, i.e. `str`
     - tuple, i.e. `(str, int)`
     - `typing.Literal['foo']`
     - `click.typing.<Foo>` (which includes click.Choice(...))
    """
    
    attrs['required'] = False
    return option(*param_decls, **attrs)


def required_opt(*param_decls, **attrs):
    """`required = True, show_default = True`.

    If `default` in `attrs` and `type` is not, or vice versa, sets one based on the other.

    Unless `default = None`, in which case `type` isn't set.

    If both are missing, `type` defaults to `str`.

    `type` can be either:
     - type, i.e. `str`
     - tuple, i.e. `(str, int)`
     - `typing.Literal['foo']`
     - `click.typing.<Foo>` (which includes click.Choice(...))
    """
    
    attrs['required'] = True
    return option(*param_decls, **attrs)



from typing import Iterable, List, Union

import re
import functools
import inspect

def return_none_if_errors(*exc):
    """If no `exc` specified, returns None on any exception.
    >>> @return_none_if_errors(ValueError)
    ... def raises(exc):
    ...     raise exc()
    >>> raises(ValueError) is None
    True
    >>> raises(TypeError)
    Traceback (most recent call last):
        ...
    TypeError
    >>> @return_none_if_errors
    ... def raises(exc):
    ...     raise exc()
    >>> raises(OverflowError) is None
    True
    """
    
    def wrap(fn):
        @functools.wraps(fn)
        def decorator(*fnargs, **fnkwargs):
            
            try:
                return fn(*fnargs, **fnkwargs)
            except exc:
                return None
        
        return decorator
    
    if not exc:
        # @return_none_if_errors()    (parens but no args)
        exc = Exception
    elif inspect.isfunction(exc[0]):
        # @return_none_if_errors    (naked)
        _fn = exc[0]
        exc = Exception
        return wrap(_fn)
    
    # @return_none_if_errors(ValueError)
    return wrap
@return_none_if_errors(ValueError, TypeError)
def safeslice(val: Union[str, int, slice]) -> slice:
    """Safe constructor for slice. Handles "2", "0:2", and ":2".
    Always returns a slice (or None if conversion fails).

    >>> mylist = ['first', 'second', 'third']
    >>> mylist[safeslice("1")] == mylist[safeslice(1)] == ['first']
    True

    >>> mylist[safeslice("0:2")] == mylist[safeslice(slice(0,2))] == ['first', 'second']
    True

    >>> mylist[safeslice(":2")] == mylist[safeslice(slice(2))] == ['first', 'second']
    True

    >>> safeslice("foo") is None
    True
    """
    
    def _to_slice(_val) -> slice:
        if isinstance(_val, slice):
            return _val
        # _stop = int(_val) + 1
        return slice(int(_val))  # may raise TypeError → None
    
    if isinstance(val, str):
        val = val.strip()
        if ':' in val:
            start, _, stop = val.partition(':')
            if start == '':  # ":2"
                start = 0
            return slice(int(start), int(stop))
    return _to_slice(val)


@return_none_if_errors(ValueError, TypeError)
def safeint(val: Union[str, int]) -> int:
    """Handles int and str ("1").
    Always returns an int (or None if cannot be used as a precise index).

    >>> mylist = ['first', 'second']
    >>> mylist[safeint("0")] == mylist[safeint(0)] == 'first'
    True

    >>> all(bad is None for bad in (safeint("0:2"), safeint(slice(0, 2)), safeint("foo")))
    True
    """
    
    if isinstance(val, str):
        val = val.strip()
    return int(val)  # may raise TypeError → None
def to_int_or_slice(val):
    """Tries converting to int, then to slice if fails.
    Finally returns None if converting to slice fails as well"""
    _int = safeint(val)
    if _int is not None:
        return _int
    _slice = safeslice(val)
    if _slice is not None:
        return _slice
    return None
def unquote(string) -> str:
    # TODO: "gallery is MOBILE only if <= $BP3" -> "gallery is MOBILE only if <=" (maybe bcz bash?)
    string = str(string)
    # content = quoted_content(string)
    # if content:
    #     return content.strip()
    # return string.strip()
    match = re.fullmatch(r'(["\'])(.*)\1', string, re.DOTALL)
    if match:  # "'hello'"
        string = match.groups()[1]
    return string.strip()
def clean(string: str) -> str:
    return unquote(string.strip())

def has_duplicates(collection) -> bool:
    return len(set(collection)) < len(collection)


def get_duplicates(collection: Iterable):
    uniq = set()
    duplicates = []
    for item in collection:
        if item in uniq:
            duplicates.append(item)
        else:
            uniq.add(item)
    return duplicates


def remove_duplicates(collection: Iterable) -> List:
    without_duplicates = []
    for item in collection:
        if item not in without_duplicates:
            without_duplicates.append(item)
    return without_duplicates


def difference(coll1: Iterable, coll2: Iterable) -> List:
    diff = []
    for item in coll1:
        if item not in coll2:
            diff.append(item)
    return diff


def intersection(coll1: Iterable, coll2: Iterable) -> List:
    inter = []
    for item in coll1:
        if item in coll2:
            inter.append(item)
    return inter



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()
packages = find_packages(exclude=["tests?", "*.tests*", "*.tests*.*", "tests*.*", 'pypi_publish.py'])
print(f'{packages = }')
setup_args = dict(name='manuals',
                  # https://packaging.python.org/tutorials/packaging-projects/
                  version='0.0.17',
                  description='',
                  long_description=long_description,
                  long_description_content_type="text/markdown",
                  license='MIT',
                  author='Gilad Barnea',
                  author_email='giladbrn@gmail.com',
                  url='https://github.com/giladbarnea/manuals',
                  packages=packages,
                  keywords=[],
                  install_requires=['more-termcolor','click', 'rich','fuzzysearch'],
                  # pip install -e .[dev]
                  extras_require={
                      'dev': ['pytest',
                              'ipdb',
                              'IPython',
                              'jupyter',
                              'semver',
                              'birdseye'
                              ]
                      },
                  # classifiers=[
                  # https://pypi.org/classifiers/
                  # 'Development Status :: 4 - Beta',
                  # 'Environment :: Console',
                  # 'Intended Audience :: Developers',
                  # "License :: OSI Approved :: MIT License",
                  # 'Operating System :: OS Independent',
                  # "Programming Language :: Python :: 3 :: Only",
                  # 'Topic :: Terminals',
                  #  ],
                  python_requires='>=3.8',
                  )
setup(**setup_args)



from typing import Union, Tuple, overload

import logging
# from igit_debug.investigate import logonreturn, logreturn
from more_termcolor import colors

# from igit.abcs import prettyrepr
from .item import MutableItem, FlowItem, LexicItem
from .options import Options, NumOptions, LexicOptions
from . import util
# from igit.util.misc import darkprint, brightyellowprint


def _input(s):
    return util.clean(input(colors.brightwhite(s)))


# (str, str) or (str, FlowItem)
AnswerTuple = Union[Tuple[str, MutableItem], Tuple[str, FlowItem], Tuple[str, LexicItem]]
Answer = Union[AnswerTuple, bool]


# @prettyrepr
class BasePrompt:
    # bool or (str, MutableItem)
    answer: Answer
    options: Options
    
    # @logonreturn('self.answer', types=True)
    def __init__(self, question: str, **kwargs):
        # noinspection PyTypeChecker
        self.answer: Answer = None
        if 'flowopts' in kwargs:
            self.options.set_flow_opts(kwargs.pop('flowopts'))
        try:
            free_input = kwargs.pop('free_input')
        except KeyError:
            free_input = False
        
        # *  keyword-choices
        self.options.set_kw_options(**kwargs)
        
        # *  Complex Prompt
        dialog_string = self.dialog_string(question, free_input=free_input)
        # question = self.dialog_string(question, options, free_input=free_input)
        key, answer = self.get_answer(dialog_string, free_input=free_input)
        # darkprint(f'{repr(self)} | key: {repr(key)}, answer: {repr(answer)}')
        # *  FlowItem Answer
        if isinstance(answer, FlowItem):
            # flow_answer: FlowItem = FlowItem(answer)
            answer.execute()
            
            if answer.DEBUG:
                # debugger had already started and had finished in answer.execute() (user 'continue'd here)
                self.answer = self.get_answer(dialog_string)
            elif answer.CONTINUE:
                self.answer: Tuple[str, FlowItem] = key, answer
            else:
                raise NotImplementedError
        else:
            # *  DIDN'T answer any flow
            
            if isinstance(answer, MutableItem) and answer.is_yes_or_no:
                # darkprint(f'{repr(self)} no flow chosen, answer is yes / no. key: {repr(key)}, answer: {repr(answer)}, options: {self.options}')
                self.answer: bool = key.lower() in ('y', 'yes')
            else:
                # darkprint(f'{repr(self)} no flow chosen, answer is not yes / no. key: {repr(key)}, answer: {repr(answer)}, options: {self.options}')
                self.answer: Tuple[str, MutableItem] = key, answer
    
    # def __repr__(self) -> str:  # uncomment if has prepr() fn (originally from @prettyrepr)
    #     return f'{self.prepr()}(answer={repr(self.answer)}, options={repr(self.options)})'
    
    def dialog_string(self, question: str, *, free_input: bool) -> str:
        strings = []
        for optkey, optval in self.options.items.items():
            strings.append(f'[{optkey}]: {optval}')
        options_str = '\n\t'.join(strings)
        question_str = f'{question}'
        # if free_input:
        #     question_str += ' (free input allowed)'
        if options_str:
            dialog_string = f'{question_str}\n\t{options_str}\n\t'
        else:
            dialog_string = question_str + '\n\t'
        if free_input:
            dialog_string += '(free input allowed)\n\t'
        return dialog_string
    
    @overload
    def get_answer(self, dialog_string: str, *, free_input=False) -> Tuple[str, Union[MutableItem, LexicItem, FlowItem]]:
        ...
    
    @overload
    def get_answer(self, dialog_string: str, *, free_input=True) -> Tuple[None, str]:
        """`free_input = True`"""
        ...
    
    # @logreturn
    def get_answer(self, dialog_string: str, *, free_input=False):
        ans_key = _input(dialog_string)
        items = self.options.items
        if ans_key not in items:
            if ans_key and free_input:
                # * Free input
                return None, ans_key
            else:
                # TODO: this doesnt account for free_input = True, but no ans_key ('')
                while ans_key not in items:
                    logging.warning(f"Unknown option: '{ans_key}'")
                    ans_key = _input(dialog_string)
        ans_value = items[ans_key]
        # this is commented out because BasePrompt init needs to check if answer.is_yes_or_no
        # if hasattr(ans_value, 'value'):
        #     ans_value = ans_value.value
        
        return ans_key, ans_value


class LexicPrompt(BasePrompt):
    options: LexicOptions
    answer: Tuple[str, LexicItem]
    
    def __init__(self, prompt: str, *options: str, **kwargs):
        self.options = LexicOptions(*options)
        super().__init__(prompt, **kwargs)


class Confirmation(LexicPrompt):
    options: LexicOptions
    answer: bool
    
    def __init__(self, prompt: str, *options: str, **kwargs):
        if 'free_input' in kwargs:
            raise ValueError(f"Confirmation cannot have free input. kwargs: {kwargs}")
        super().__init__(prompt, *options, **kwargs)


class Action(BasePrompt):
    options: LexicOptions
    answer: Tuple[str, LexicItem]
    
    def __init__(self, question: str, *actions: str, **kwargs):
        if not actions:
            raise ValueError(f'At least one action is required')
        self.options = LexicOptions(*actions)
        if self.options.any_item(lambda item: item.is_yes_or_no):
            raise ValueError(f"Actions cannot include a 'yes' or 'no'. Received: {repr(actions)}")
        super().__init__(question, **kwargs)


# TODO: Choice can return index! change docs
class Choice(BasePrompt):
    """If `free_input=True`, `answer` may be (None, str) or (None, slice) (if input is e.g. "2:5").
    
    Otherwise, if a numeric choice is made, `answer` is (slice, MutableItem).
    
    If a lexic choice is made, `answer` is (key: str, LexicItem)."""
    options: NumOptions
    answer: Union[Tuple[slice, MutableItem], Tuple[str, LexicItem], Tuple[None, str], Tuple[None, slice]]
    
    def get_answer(self, question: str, *, free_input=False) -> Tuple[Union[None, str, slice], Union[MutableItem, LexicItem, str, slice]]:
        ans_key, ans_value = super().get_answer(question, free_input=free_input)
        if ans_key is None:
            # free input
            indexer = util.to_int_or_slice(ans_value)
            if indexer is not None:
                ans_value = indexer
        else:
            indexer = util.to_int_or_slice(ans_key)
            if indexer is not None:
                ans_key = indexer
        return ans_key, ans_value
    
    def __init__(self, question: str, *options: str, **kwargs):
        if not options:
            raise ValueError(f'At least one option is required when using Choice (contrary to Prompt)')
        self.options = NumOptions(*options)
        super().__init__(question, **kwargs)


def generic(prompt: str, *options: str, **kwargs: Union[str, tuple, bool]):
    """Most permissive, a simple wrapper for LexicPrompt. `options` are optional, `kwargs` are optional.
    Examples::

        generic('This and that, continue?', 'yes', flowopts='quit', free_input=True) → [y], [q] (free input allowed)
    """
    
    return LexicPrompt(prompt, *options, **kwargs).answer


@overload
def choose(prompt, *options: str, **kwargs: Union[str, tuple, bool]) -> Tuple[None, slice]:
    """`free_input = True` and user input was e.g. "2:5" thus converted successfully to slice."""
    ...


@overload
def choose(prompt, *options: str, **kwargs: Union[str, tuple, bool]) -> Tuple[None, str]:
    """`free_input = True` and user input was some string."""
    ...


@overload
def choose(prompt, *options: str, **kwargs: Union[str, tuple, bool]) -> Tuple[slice, MutableItem]:
    """User chose a numeric selection that was converted to slice."""
    ...


@overload
def choose(prompt, *options: str, **kwargs: Union[str, tuple, bool]) -> Tuple[str, LexicItem]:
    """User chose a lexic selection."""
    ...


def choose(prompt, *options: str, **kwargs: Union[str, tuple, bool]):
    """Presents `options` by *index*. Expects at least one option.
    
    If `free_input=True`, return type may be (None, str) or (None, slice) (if input is e.g. "2:5").
    
    Otherwise, if a numeric choice is made, return type is (idx: int, MutableItem).
    
    If a lexic choice is made, return type is (key: str, LexicItem).
    """
    
    answer = Choice(prompt, *options, **kwargs).answer
    return answer


# TODO: implement prompt.confirm('yes', 'quit', no='open with current') that returns bool (see search.py _choose_from_many())
def confirm(prompt, **kwargs: Union[str, tuple, bool]) -> bool:
    """A 'y/n' prompt.

    If `options` contains any "special options", they are presented by key.
    Examples::

        confirm('ice cream?', flowopts='quit') → [y], [n], [q]
        confirm('pizza?', flowopts=True) → [y], [n], [c], [d], [q]
        confirm('burger?', flowopts=('quit', 'debug')) → [y], [n], [q], [d]
        confirm('proceed?') → [y], [n]
    """
    
    return Confirmation(prompt, 'yes', 'no', **kwargs).answer


def action(question, *actions, **kwargs: Union[str, tuple, bool]) -> Tuple[str, LexicItem]:
    """Presents `options` by *key*.
    Compared to `confirm()`, `action()` can't be used to prompt for yes/no but instead prompts for strings.
    Example::

        action('uncommitted changes', 'stash & apply', flowopts = True)

    :param actions: must not be empty, and cannot contain 'yes' or 'no'.
    :param special_opts:
        If True, special options: ('continue', 'debug', 'quit') are also presented.
        If a str, it has to be one of the special options above.
        If tuple, has to contain only special options above.
    """
    
    return Action(question, *actions, **kwargs).answer
